<?php /*

 Composr
 Copyright (c) Christopher Graham, 2004-2024

 See docs/LICENSE.md for full licensing information.


 NOTE TO PROGRAMMERS:
   Do not edit this file. If you need to make changes, save your changed file to the appropriate *_custom folder
   **** If you ignore this advice, then your website upgrades (e.g. for bug fixes) will likely kill your changes ****

*/

/**
 * @license    http://opensource.org/licenses/cpal_1.0 Common Public Attribution License
 * @copyright  Christopher Graham
 * @package    banners
 */

/**
 * Module page class.
 */
class Module_admin_banners
{
    /**
     * Find details of the module.
     *
     * @return ?array Map of module info (null: module is disabled)
     */
    public function info() : ?array
    {
        $info = [];
        $info['author'] = 'Chris Graham';
        $info['organisation'] = 'Composr';
        $info['hacked_by'] = null;
        $info['hack_version'] = null;
        $info['version'] = 2;
        $info['locked'] = false;
        $info['min_cms_version'] = 11.0;
        $info['addon'] = 'banners';
        return $info;
    }

    /**
     * Find entry-points available within this module.
     *
     * @param  boolean $check_perms Whether to check permissions
     * @param  ?MEMBER $member_id The member to check permissions as (null: current user)
     * @param  boolean $support_crosslinks Whether to allow cross links to other modules (identifiable via a full-page-link rather than a screen-name)
     * @param  boolean $be_deferential Whether to avoid any entry-point (or even return null to disable the page in the Sitemap) if we know another module, or page_group, is going to link to that entry-point. Note that "!" and "browse" entry points are automatically merged with container page nodes (likely called by page-groupings) as appropriate.
     * @return ?array A map of entry points (screen-name=>language-code/string or screen-name=>[language-code/string, icon-theme-image]) (null: disabled)
     */
    public function get_entry_points(bool $check_perms = true, ?int $member_id = null, bool $support_crosslinks = true, bool $be_deferential = false) : ?array
    {
        if (!addon_installed('banners')) {
            return null;
        }

        return [
            'browse' => ['BANNER_STATISTICS', 'menu/cms/banners'],
        ];
    }

    public $title;

    /**
     * Module pre-run function. Allows us to know metadata for <head> before we start streaming output.
     *
     * @return ?Tempcode Tempcode indicating some kind of exceptional output (null: none)
     */
    public function pre_run() : ?object
    {
        $error_msg = new Tempcode();
        if (!addon_installed__messaged('banners', $error_msg)) {
            return $error_msg;
        }

        $type = get_param_string('type', 'browse');

        require_lang('banners');
        require_code('banners');

        if ($type == 'browse') {
            $also_url = build_url(['page' => 'cms_banners'], get_module_zone('cms_banners'));
            attach_message(do_lang_tempcode('menus:ALSO_SEE_ADMIN', escape_html($also_url->evaluate())), 'inform', true);

            $this->title = get_screen_title('BANNER_STATISTICS');
        }

        return null;
    }

    /**
     * Execute the module.
     *
     * @return Tempcode The result of execution
     */
    public function run() : object
    {
        $type = get_param_string('type', 'browse');

        if ($type == 'browse') {
            return $this->banner_statistics();
        }

        return new Tempcode();
    }

    /**
     * The UI to show a results table of banner details/statistics.
     *
     * @return Tempcode The UI
     */
    public function banner_statistics() : object
    {
        require_lang('banners');
        check_privilege('view_anyones_banner_stats');

        $start = get_param_integer('start', 0);
        $max = get_param_integer('max', 50);
        $sortables = ['name' => do_lang_tempcode('NAME'), 'add_date' => do_lang_tempcode('ADDED')];
        $current_sorting = get_param_string('sort', 'name ASC', INPUT_FILTER_GET_COMPLEX);
        list($sql_sort, $sort_order, $sortable) = process_sorting_params('banner', $current_sorting);

        $__sum = $GLOBALS['SITE_DB']->query_select_value('banners', 'SUM(views_from)');
        $_sum = @intval($__sum);
        $has_banner_network = ($_sum != 0);

        require_code('templates_results_table');
        $_header_row = [do_lang_tempcode('NAME'), do_lang_tempcode('TYPE'), do_lang_tempcode('BANNER_TYPE')];
        $filtercode = ['name<name_op><name>', 'deployment_agreement=<deployment_agreement>', 'b_type=<b_type>'];
        $filtercode_labels = ['deployment_agreement=' . do_lang('TYPE'), 'b_type=' . do_lang('BANNER_TYPE')];
        $filtercode_types = ['b_type=list'];
        if ($has_banner_network) {
            $_header_row = array_merge($_header_row, [do_lang_tempcode('BANNER_HITS_FROM'), do_lang_tempcode('BANNER_VIEWS_FROM')]);
            $filtercode[] = 'hits_from<hits_from_op><hits_from>';
            $filtercode[] = 'views_from<views_from_op><views_from>';
        }
        $_header_row = array_merge($_header_row, [do_lang_tempcode('BANNER_HITS_TO'), do_lang_tempcode('BANNER_VIEWS_TO'), do_lang_tempcode('BANNER_CLICKTHROUGH'), do_lang_tempcode('DISPLAY_LIKELIHOOD'), do_lang_tempcode('SUBMITTER'), do_lang_tempcode('ADDED')]);
        $filtercode[] = 'hits_to<hits_to_op><hits_to>';
        $filtercode[] = 'views_to<views_to_op><views_to>';
        $filtercode[] = 'display_likelihood<display_likelihood_op><display_likelihood>';
        $filtercode[] = 'submitter=<submitter>';
        $filtercode[] = 'add_date<add_date_op><add_date>';
        $filtercode_labels[] = 'add_date=' . do_lang('ADDED');
        if (addon_installed('validation')) {
            $_header_row[] = protect_from_escaping(do_template('COMCODE_ABBR', ['_GUID' => '7a2ed997384b823b25dc3c70d4ff82d6', 'TITLE' => do_lang_tempcode('VALIDATED'), 'CONTENT' => do_lang_tempcode('VALIDATED_SHORT')]));
            $filtercode[] = 'validated=<validated>';
        }
        $header_row = results_header_row($_header_row, $sortables, 'sort', $sortable . ' ' . $sort_order);

        // Prepare Filtercode
        require_code('filtercode');
        $active_filters = get_params_filtercode();

        // Build WHERE query from Filtercode
        $where = [];
        list($extra_join, $filtercode_end) = filtercode_to_sql($GLOBALS['SITE_DB'], parse_filtercode($active_filters), 'banner', null);
        $end = ' ORDER BY ' . $sql_sort;

        $rows = $GLOBALS['SITE_DB']->query_select('banners r', ['*'], $where, $filtercode_end . $end, $max, $start);
        $max_rows = $GLOBALS['SITE_DB']->query_select_value('banners r', 'COUNT(*)', $where, $filtercode_end);

        // Construct the table
        $result_entries = new Tempcode();
        foreach ($rows as $myrow) {
            $name = hyperlink(build_url(['page' => 'banners', 'type' => 'view', 'source' => $myrow['name']], get_module_zone('banners')), $myrow['name'], false, true);

            switch ($myrow['deployment_agreement']) {
                case BANNER_PERMANENT:
                    $type = do_lang_tempcode('BANNER_PERMANENT');
                    break;
                case BANNER_CAMPAIGN:
                    $type = do_lang_tempcode('BANNER_HITS_LEFT', do_lang_tempcode('BANNER_CAMPAIGN'), escape_html(integer_format($myrow['campaign_remaining'], 0)));
                    break;
                case BANNER_FALLBACK:
                    $type = do_lang_tempcode('BANNER_FALLBACK');
                    break;
                default:
                    $type = do_lang_tempcode('UNKNOWN');
            }

            $banner_type = $myrow['b_type'];
            if ($banner_type == '') {
                $banner_type = do_lang('_DEFAULT');
            }

            $date = get_timezoned_date($myrow['add_date']);

            $hits_from = integer_format($myrow['hits_from']);
            $views_from = integer_format($myrow['views_from']);
            $hits_to = ($myrow['site_url'] == '') ? do_lang_tempcode('CANT_TRACK') : protect_from_escaping(escape_html(integer_format($myrow['hits_to'], 0)));
            $views_to = ($myrow['site_url'] == '') ? do_lang_tempcode('CANT_TRACK') : protect_from_escaping(escape_html(integer_format($myrow['views_to'], 0)));

            if ($myrow['views_to'] != 0) {
                $click_through = protect_from_escaping(escape_html(float_format(round(100.0 * ($myrow['hits_to'] / $myrow['views_to']))) . '%'));
            } else {
                $click_through = do_lang_tempcode('NA_EM');
            }

            $username = $GLOBALS['FORUM_DRIVER']->member_profile_hyperlink($myrow['submitter']);

            $display_likelihood = $myrow['display_likelihood'];

            $validated = ($myrow['validated'] == 1) ? do_lang('YES') : do_lang('NO');
            if (($myrow['expiry_date'] !== null) && ($myrow['expiry_date'] < time())) {
                $validated .= do_lang('BUT_EXPIRED');
            }

            $result = [$name, $type, escape_html($banner_type)];
            if ($has_banner_network) {
                $result = array_merge($result, [escape_html($hits_from), escape_html($views_from)]);
            }
            $result = array_merge($result, [$hits_to, $views_to, $click_through, escape_html(strval($display_likelihood)), $username, escape_html($date)]);
            if (addon_installed('validation')) {
                $result[] = escape_html($validated);
            }

            $result_entries->attach(results_entry($result, true));
        }

        $table = results_table(do_lang_tempcode('BANNERS'), $start, 'start', $max, 'max', $max_rows, $header_row, $result_entries, $sortables, $sortable, $sort_order, 'sort');

        $filtercode_box = do_block('main_content_filtering', [
            'param' => implode(',', $filtercode),
            'content_type' => 'banner',
            'labels' => implode(',', $filtercode_labels),
            'types' => implode(',', $filtercode_types),
        ]);

        $tpl = do_template('RESULTS_TABLE_SCREEN', [
            '_GUID' => '82133b12ecdece7cd22f879c0c20ff44',
            'RESULTS_TABLE' => $table,
            'TITLE' => $this->title,
            'FILTERCODE_BOX' => $filtercode_box,
        ]);

        require_code('templates_internalise_screen');
        return internalise_own_screen($tpl);
    }
}
